/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*   This file has been modified from the Seban Giuroiu  and Wei-Keng Liao   */
/*   version                                                                 */
/*   to support kmeans++ and kmeans++ with point retraction initialization   */
/*   of centroids                                                            */
/*   Incorporated by Isaac Goicovich, Paulo Olivares and Cecilia Hernandez   */
/*   We modify the code by Serban Giuroiu and Wei-Keng Liao beacuase their   */
/*   statement given below.                                                  */
/*   ------------------------------------------------------------------------*/
/*   File:         cuda_kmeans.cu  (CUDA version)                            */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// Copyright (c) 2005 Wei-keng Liao
// Copyright (c) 2011 Serban Giuroiu
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// -----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <curand.h>
#include <curand_kernel.h>

#include "kmeans.h"


__global__ void dist_euclideana(float *in_data,float *out_data, unsigned int pin,unsigned int tam){
    int hId = (blockDim.x*blockIdx.x + threadIdx.x);
    int hId3=hId*3;
    float p1=in_data[pin],p2=in_data[pin+tam],p3=in_data[pin+2*tam];
      if( hId < tam ){
           float v1=in_data[hId]-p1;
           float v2=in_data[hId+tam]-p2;
           float v3=in_data[hId+tam*2]-p3;
           out_data[hId]= (v1)*(v1)+(v2)*(v2)+(v3)*(v3) ; 
	}
    }

__global__ void red50(float *g_idata, float *g_odata, unsigned int *p_odata, unsigned int n,float *puntos_ini,unsigned int index,float *puntos,unsigned int numClusters,int tam)
{
  __shared__ float sdata[1024*2];
  unsigned int blockSize=blockDim.x;
  unsigned int tid = threadIdx.x;
  
  unsigned int tid2=tid+blockSize;

  
  sdata[tid]=g_idata[tid];sdata[tid2]=p_odata[tid]; 
  __syncthreads();
  if (tid < 512) { if(sdata[tid + 512]>sdata[tid]) { sdata[tid]=sdata[tid + 512]; sdata[tid2]=sdata[tid2+512]; } } __syncthreads(); 
  if (tid < 256) { if(sdata[tid + 256]>sdata[tid]) { sdata[tid]=sdata[tid + 256]; sdata[tid2]=sdata[tid2+256]; } } __syncthreads(); 
  if (tid < 128) { if(sdata[tid + 128]>sdata[tid]) { sdata[tid]=sdata[tid + 128]; sdata[tid2]=sdata[tid2+128]; } } __syncthreads(); 
  if (tid < 64) { if(sdata[tid + 64]>sdata[tid]) { sdata[tid]=sdata[tid + 64]; sdata[tid2]=sdata[tid2+64]; }  } __syncthreads(); 
  
  if (tid < 32) {
    if(sdata[tid + 32]>sdata[tid]) { sdata[tid]=sdata[tid + 32]; sdata[tid2]=sdata[tid2+32]; } 
    if(sdata[tid + 16]>sdata[tid]) { sdata[tid]=sdata[tid + 16]; sdata[tid2]=sdata[tid2+16]; } 
    if(sdata[tid + 8]>sdata[tid]) { sdata[tid]=sdata[tid + 8]; sdata[tid2]=sdata[tid2+8]; } 
    if(sdata[tid + 4]>sdata[tid]) { sdata[tid]=sdata[tid + 4]; sdata[tid2]=sdata[tid2+4]; } 
    if(sdata[tid + 2]>sdata[tid]) { sdata[tid]=sdata[tid + 2]; sdata[tid2]=sdata[tid2+2]; } 
    if(sdata[tid + 1]>sdata[tid]) { sdata[tid]=sdata[tid + 1]; sdata[tid2]=sdata[tid2+1]; } 
  }

  __syncthreads();

    if (tid == 0) {g_odata[0] = sdata[0];}
    if (tid == 1) {p_odata[0] = sdata[1024];}
    if (tid == 2) {puntos_ini[index]=puntos[ (unsigned int)sdata[1024] ];}
    if (tid == 3) {puntos_ini[index+numClusters]=puntos[ (unsigned int)sdata[1024]+tam ];}
    if (tid == 4) {puntos_ini[index+2*numClusters]=puntos[ (unsigned int)sdata[1024]+2*tam ];}
    
}

__global__ void red51(float *g_idata, float *g_odata, unsigned int *p_odata, unsigned int n,float *puntos)
{
  __shared__ float sdata[1024*2];
  unsigned int blockSize=blockDim.x;
  unsigned int bID = blockIdx.x;
  unsigned int tid = threadIdx.x;
  unsigned int i = bID*(blockSize) + tid;
  unsigned int gridSize = blockSize*gridDim.x;

  unsigned int tid2=tid+blockSize;
  if(i<n){sdata[tid] = g_idata[i]; sdata[tid2]=i;}
  __syncthreads();
  i += gridSize;
 while (i < n) { if( g_idata[i]>sdata[tid] ) { sdata[tid]=g_idata[i] ; sdata[tid2]=i; }  ; i += gridSize; } 

  __syncthreads();

  if (tid < 512) { if(sdata[tid + 512]>sdata[tid]) { sdata[tid]=sdata[tid + 512]; sdata[tid2]=sdata[tid2+512]; } } __syncthreads(); 
  if (tid < 256) { if(sdata[tid + 256]>sdata[tid]) { sdata[tid]=sdata[tid + 256]; sdata[tid2]=sdata[tid2+256]; } } __syncthreads(); 
  if (tid < 128) { if(sdata[tid + 128]>sdata[tid]) { sdata[tid]=sdata[tid + 128]; sdata[tid2]=sdata[tid2+128]; } } __syncthreads(); 
  if (tid < 64) { if(sdata[tid + 64]>sdata[tid]) { sdata[tid]=sdata[tid + 64]; sdata[tid2]=sdata[tid2+64]; }  } __syncthreads(); 
  
  if (tid < 32) {
     if(sdata[tid + 32]>sdata[tid]) { sdata[tid]=sdata[tid + 32]; sdata[tid2]=sdata[tid2+32]; } 
     if(sdata[tid + 16]>sdata[tid]) { sdata[tid]=sdata[tid + 16]; sdata[tid2]=sdata[tid2+16]; } 
     if(sdata[tid + 8]>sdata[tid]) { sdata[tid]=sdata[tid + 8]; sdata[tid2]=sdata[tid2+8]; } 
     if(sdata[tid + 4]>sdata[tid]) { sdata[tid]=sdata[tid + 4]; sdata[tid2]=sdata[tid2+4]; } 
     if(sdata[tid + 2]>sdata[tid]) { sdata[tid]=sdata[tid + 2]; sdata[tid2]=sdata[tid2+2]; } 
     if(sdata[tid + 1]>sdata[tid]) { sdata[tid]=sdata[tid + 1]; sdata[tid2]=sdata[tid2+1]; } 
  }

  __syncthreads();

    if (tid == 0) {g_odata[bID] = sdata[tid]; p_odata[bID]=sdata[tid2];} 
    
}

 __global__ void next_dist(float *in_data, float *out_distancias,unsigned int* in_pin ,unsigned int tam){
    int hId = blockIdx.x*blockDim.x + threadIdx.x;
    unsigned int memo3=in_pin[0];
    float new_dist;
    float p1=in_data[memo3],p2=in_data[memo3+tam],p3=in_data[memo3+tam*2];
    if( hId < tam ){
        float f1=(in_data[hId]-p1);
        float f2=(in_data[hId+tam]-p2);
        float f3=(in_data[hId+tam*2]-p3);
        new_dist = f1*f1+f2*f2+f3*f3;
        
        if(new_dist < out_distancias[hId]){
            out_distancias[hId] = new_dist;
        }
    }
}


__host__ __device__ static inline int nextPowerOfTwo(int n) {
    n--;

    n = n >>  1 | n;
    n = n >>  2 | n;
    n = n >>  4 | n;
    n = n >>  8 | n;
    n = n >> 16 | n;
//  n = n >> 32 | n;    //  For 64-bit ints

    return ++n;
}

/*----< euclid_dist_2() >----------------------------------------------------*/
/* square of Euclid distance between two multi-dimensional points            */
__host__ __device__ inline static
float euclid_dist_2(int    numCoords,
                    int    numObjs,
                    int    numClusters,
                    float *objects,     // [numCoords][numObjs]
                    float *clusters,    // [numCoords][numClusters]
                    int    objectId,
                    int    clusterId)
{
    int i;
    float ans=0.0;

    for (i = 0; i < numCoords; i++) {
        ans += (objects[numObjs * i + objectId] - clusters[numClusters * i + clusterId]) *
               (objects[numObjs * i + objectId] - clusters[numClusters * i + clusterId]);
    }

    return(ans);
}


/*----< find_nearest_cluster() >---------------------------------------------*/
__global__ static
void find_nearest_cluster(int numCoords,
                          int numObjs,
                          int numClusters,
                          float *objects,           //  [numCoords][numObjs]
                          float *deviceClusters,    //  [numCoords][numClusters]
                          int *membership,          //  [numObjs]
                          int *intermediates)
{
    extern __shared__ char sharedMemory[];

    //  The type chosen for membershipChanged must be large enough to support
    //  reductions! There are blockDim.x elements, one for each thread in the
    //  block. See numThreadsPerClusterBlock in cuda_kmeans().
    unsigned char *membershipChanged = (unsigned char *)sharedMemory;
#if BLOCK_SHARED_MEM_OPTIMIZATION
    float *clusters = (float *)(sharedMemory + blockDim.x);
#else
    float *clusters = deviceClusters;
#endif

    membershipChanged[threadIdx.x] = 0;
#if BLOCK_SHARED_MEM_OPTIMIZATION
    //  BEWARE: We can overrun our shared memory here if there are too many
    //  clusters or too many coordinates! For reference, a Tesla C1060 has 16
    //  KiB of shared memory per block, and a GeForce GTX 480 has 48 KiB of
    //  shared memory per block.
    for (int i = threadIdx.x; i < numClusters; i += blockDim.x) {
        for (int j = 0; j < numCoords; j++) {
            clusters[numClusters * j + i] = deviceClusters[numClusters * j + i];
        }
    }
    __syncthreads();
#endif

    int objectId = blockDim.x * blockIdx.x + threadIdx.x;

    if (objectId < numObjs) {
        int   index, i;
        float dist, min_dist;

        /* find the cluster id that has min distance to object */
        index    = 0;
        min_dist = euclid_dist_2(numCoords, numObjs, numClusters,
                                 objects, clusters, objectId, 0);

        for (i=1; i<numClusters; i++) {
            dist = euclid_dist_2(numCoords, numObjs, numClusters,
                                 objects, clusters, objectId, i);
            /* no need square root */
            if (dist < min_dist) { /* find the min and its array index */
                min_dist = dist;
                index    = i;
            }
        }

        if (membership[objectId] != index) {
            membershipChanged[threadIdx.x] = 1;
        }

        /* assign the membership to object objectId */
        membership[objectId] = index;

        __syncthreads();    //  For membershipChanged[]

       //  blockDim.x *must* be a power of two!
        for (unsigned int s = blockDim.x / 2; s > 0; s >>= 1) {
            if (threadIdx.x < s) {
                membershipChanged[threadIdx.x] +=
                    membershipChanged[threadIdx.x + s];
            }
            __syncthreads();
        }

        if (threadIdx.x == 0) {
            intermediates[blockIdx.x] = membershipChanged[0];
        }
    }
}

__global__ static
void compute_delta(int *deviceIntermediates,
                   int numIntermediates,    //  The actual number of intermediates
                   int numIntermediates2)   //  The next power of two
{
    //  The number of elements in this array should be equal to
    //  numIntermediates2, the number of threads launched. It *must* be a power
    //  of two!
    extern __shared__ unsigned int intermediates[];

    //  Copy global intermediate values into shared memory.
    intermediates[threadIdx.x] =
        (threadIdx.x < numIntermediates) ? deviceIntermediates[threadIdx.x] : 0;

    __syncthreads();

    //  numIntermediates2 *must* be a power of two!
    for (unsigned int s = numIntermediates2 / 2; s > 0; s >>= 1) {
        if (threadIdx.x < s) {
            intermediates[threadIdx.x] += intermediates[threadIdx.x + s];
        }
        __syncthreads();
    }

    if (threadIdx.x == 0) {
        deviceIntermediates[0] = intermediates[0];
    }
}


int *cuda_kmeans(float *objects,      /* in: [numObjs][numCoords] */
                   int     numCoords,    /* no. features */
                   int     numObjs,      /* no. objects */
                   int     numClusters,  /* no. clusters */
                   float   threshold,    /* % objects change membership */
                   int    *loop_iterations,
		   float *matres)
{
    int      i, j, index, loop=0;
    int     *newClusterSize; /* [numClusters]: no. objects assigned in each
                                new cluster */
    float    delta;          /* % of objects change their clusters */
    float  **dimObjects;
    float  **clusters;       /* out: [numClusters][numCoords] */
    float  **dimClusters;
    float  **newClusters;    /* [numCoords][numClusters] */
    int *membership;

    float *deviceObjects;
    float *deviceClusters;
    int *deviceMembership;
    int *deviceIntermediates;
    float  *devdimClusters;

    membership = (int*) malloc(numObjs * sizeof(int));

    checkCuda(cudaMalloc(&devdimClusters, numClusters*numCoords*sizeof(float)));

    //  Copy objects given in [numObjs][numCoords] layout to new
    //  [numCoords][numObjs] layout

    malloc2D(dimObjects, numCoords, numObjs, float);
    for (i = 0; i < numCoords; i++) {
        for (j = 0; j < numObjs; j++) {
            //dimObjects[i][j] = objects[j][i];
            dimObjects[i][j] = objects[j*numCoords + i];
        }
    }

    checkCuda(cudaMalloc(&deviceObjects, numObjs*numCoords*sizeof(float)));
    checkCuda(cudaMemcpy(deviceObjects, dimObjects[0],
              numObjs*numCoords*sizeof(float), cudaMemcpyHostToDevice));

    /* pick first numClusters elements of objects[] as initial cluster centers*/
    malloc2D(dimClusters, numCoords, numClusters, float);

    srand (time(NULL));
    //srand (101);
    int init_rand = rand()%numObjs;
    dimClusters[0][0] = dimObjects[0][init_rand];
    dimClusters[1][0] = dimObjects[1][init_rand];
    dimClusters[2][0] = dimObjects[2][init_rand];
   
    float px0 = dimClusters[0][0];
    float py0 = dimClusters[1][0];
    float pz0 = dimClusters[2][0];

    float *d_distancias;
    float *distancias;
    distancias = (float*) malloc(numObjs*3 * sizeof(float));
    unsigned int *d_posiciones;
    float *d_aux;
    unsigned int numObjects=3*numObjs;
    unsigned int nextpow2=nextPowerOfTwo(numObjects);


    cudaMalloc(&d_distancias, numObjects*sizeof(float));
    cudaMalloc(&d_posiciones, numObjects*sizeof(unsigned int));
    cudaMalloc(&d_aux, nextpow2*sizeof(float));


    unsigned int mtpb=1024;
    unsigned int numg_de=1024; 


    double ptime = wtime();
    dist_euclideana<<<numg_de,mtpb >>>(deviceObjects,d_distancias,init_rand,numObjs);
    checkCuda(cudaDeviceSynchronize());

    for(unsigned int j=0;j<(numClusters-1);j++){
      red51<<< mtpb,mtpb >>>(d_distancias,d_aux,d_posiciones,numObjs,deviceObjects);
      red50<<< 1,mtpb >>>(d_aux,d_aux,d_posiciones,mtpb,devdimClusters,(j+1),deviceObjects,numClusters,numObjs);
      next_dist<<<numg_de,mtpb >>>(deviceObjects,d_distancias,d_posiciones,numObjs);
    }
    cudaDeviceSynchronize();

    cudaMemcpy(dimClusters[0], devdimClusters, numClusters*sizeof(float), cudaMemcpyDeviceToHost);
    cudaMemcpy(dimClusters[1], devdimClusters+numClusters, numClusters*sizeof(float), cudaMemcpyDeviceToHost);
    cudaMemcpy(dimClusters[2], devdimClusters+numClusters*2, numClusters*sizeof(float), cudaMemcpyDeviceToHost);

    double eptime = wtime();
    //printf(" kmeanspp time %f\t", eptime - ptime);
    //printf(" save kpp time %f\n", wtime() - eptime);

    dimClusters[0][0] = px0;
    dimClusters[1][0] = py0;
    dimClusters[2][0] = pz0;
    float h_mpoint[3], s_mpoint[3] = {0.0,0.0,0.0};
    for (i = 0; i < numClusters; i++) {
	s_mpoint[0] += dimClusters[0][i];
	s_mpoint[1] += dimClusters[1][i];
	s_mpoint[2] += dimClusters[2][i];
    }
    h_mpoint[0] = s_mpoint[0]/numClusters;
    h_mpoint[1] = s_mpoint[1]/numClusters;
    h_mpoint[2] = s_mpoint[2]/numClusters;
    // retraction of 5 percent (0.05)
    float retract = 0.05;
    float retractComp = 1.0-retract;
    for (i = 0; i < numClusters; i++) {
       for (j = 0; j < numCoords; j++) {
          dimClusters[j][i] = dimClusters[j][i]*retractComp + h_mpoint[j]*retract;
       }
    }

    //printf("\nh_mpoint\n[%f , %f , %f] \n\n",h_mpoint[0],h_mpoint[1],h_mpoint[2] );
    cudaFree(d_distancias);
    cudaFree(d_aux);
    cudaFree(d_posiciones);



    /* initialize membership[] */
    for (i=0; i<numObjs; i++) membership[i] = -1;

    /* need to initialize newClusterSize and newClusters[0] to all 0 */
    newClusterSize = (int*) calloc(numClusters, sizeof(int));
    assert(newClusterSize != NULL);

    malloc2D(newClusters, numCoords, numClusters, float);

    //  To support reduction, numThreadsPerClusterBlock *must* be a power of
    //  two, and it *must* be no larger than the number of bits that will
    //  fit into an unsigned char, the type used to keep track of membership
    //  changes in the kernel.
    const unsigned int numThreadsPerClusterBlock = 1024; // este es para todas las fibras punto1-all
    const unsigned int numClusterBlocks =
        (numObjs + numThreadsPerClusterBlock - 1) / numThreadsPerClusterBlock;
#if BLOCK_SHARED_MEM_OPTIMIZATION
    const unsigned int clusterBlockSharedDataSize =
        numThreadsPerClusterBlock * sizeof(unsigned char) +
        numClusters * numCoords * sizeof(float);

    cudaDeviceProp deviceProp;
    int deviceNum;
    cudaGetDevice(&deviceNum);
    cudaGetDeviceProperties(&deviceProp, deviceNum);

    if (clusterBlockSharedDataSize > deviceProp.sharedMemPerBlock) {
        err("WARNING: Your CUDA hardware has insufficient block shared memory. "
            "You need to recompile with BLOCK_SHARED_MEM_OPTIMIZATION=0. "
            "See the README for details.\n");
    }
#else
    const unsigned int clusterBlockSharedDataSize =
        numThreadsPerClusterBlock * sizeof(unsigned char);
#endif

    const unsigned int numReductionThreads =
        nextPowerOfTwo(numClusterBlocks);
    const unsigned int reductionBlockSharedDataSize =
        numReductionThreads * sizeof(unsigned int);

    checkCuda(cudaMalloc(&deviceClusters, numClusters*numCoords*sizeof(float)));
    checkCuda(cudaMalloc(&deviceMembership, numObjs*sizeof(int)));
    checkCuda(cudaMalloc(&deviceIntermediates, numReductionThreads*sizeof(unsigned int)));

    checkCuda(cudaMemcpy(deviceMembership, membership, numObjs*sizeof(int), cudaMemcpyHostToDevice));

    do {
        checkCuda(cudaMemcpy(deviceClusters, dimClusters[0],
                  numClusters*numCoords*sizeof(float), cudaMemcpyHostToDevice));

        find_nearest_cluster<<< numClusterBlocks, numThreadsPerClusterBlock, clusterBlockSharedDataSize >>> (numCoords, numObjs, numClusters, deviceObjects, deviceClusters, deviceMembership, deviceIntermediates);

        cudaDeviceSynchronize(); 

        compute_delta <<< 1, numReductionThreads, reductionBlockSharedDataSize >>> (deviceIntermediates, numClusterBlocks, numReductionThreads);

        cudaDeviceSynchronize(); 

        int d;
        checkCuda(cudaMemcpy(&d, deviceIntermediates,
                  sizeof(int), cudaMemcpyDeviceToHost));
        delta = (float)d;
        checkCuda(cudaMemcpy(membership, deviceMembership,
                  numObjs*sizeof(int), cudaMemcpyDeviceToHost));

        for (i=0; i<numObjs; i++) {
            /* find the array index of nestest cluster center */
            index = membership[i];

            /* update new cluster centers : sum of objects located within */
            newClusterSize[index]++;
            for (j=0; j<numCoords; j++) newClusters[j][index] += objects[i*numCoords + j];
        }

        //  TODO: Flip the nesting order
        //  TODO: Change layout of newClusters to [numClusters][numCoords]
        /* average the sum and replace old cluster centers with newClusters */
        for (i=0; i<numClusters; i++) {
            for (j=0; j<numCoords; j++) {
                if (newClusterSize[i] > 0)
                    dimClusters[j][i] = newClusters[j][i] / newClusterSize[i];
                newClusters[j][i] = 0.0;   /* set back to 0 */
            }
            newClusterSize[i] = 0;   /* set back to 0 */
        }

        delta /= numObjs;
    } while (delta > threshold && loop++ < 500);
   *loop_iterations = loop + 1;

    /* allocate a 2D space for returning variable clusters[] (coordinates
       of cluster centers) */
    
    for (i = 0; i < numClusters; i++) {
        for (j = 0; j < numCoords; j++) {
            matres[i*numCoords + j] = dimClusters[j][i];
        }
    }

  
    cudaFree(deviceObjects);
    cudaFree(deviceClusters);
    cudaFree(deviceIntermediates);

    free(dimObjects[0]);
    free(dimObjects);
    free(dimClusters[0]);
    free(dimClusters);
    free(newClusters[0]);
    free(newClusters);
    free(newClusterSize);

    free(membership);
    return deviceMembership;
}



