#include <bits/stdc++.h>
#include <thrust/sort.h>
#include <thrust/host_vector.h>
#include <thrust/device_vector.h>
#include <sys/time.h>
#include <thrust/extrema.h>
#include <thrust/execution_policy.h>
#include <thrust/reduce.h>
#include <thrust/transform.h>
#include <thrust/functional.h>
#include <sys/stat.h>
#include "kmeans.h"
#include "Algorithm.h"
#include "AdjacencyListAlgorithm.h"
#include <string>
#include <queue>
#include <omp.h>


using namespace std;

// default values for dRmax and dMmax
float dRmax = 6.0;
float dMmax = 6.0;

//Array for Kp values  
int Kp[21];
	
//To init the Kp array, by default the values are 200 for internal points and 300 for extremes points (0 and 20)
void init(){
	for(int i=0;i<21;i++){
		Kp[i]=0;
		if(i==0 || i==3 || i==10 || i==17 || i==20){
			if(i==0 || i==20) Kp[i] = 300;
			else Kp[i] = 200;
		}
	}
}

// To compute time
double wtime(void) {
	double          now_time;
	struct timeval  etstart;
	struct timezone tzp;

	if (gettimeofday(&etstart, &tzp) == -1)
		perror("Error: calling gettimeofday() not successful.\n");

	now_time = ((double)etstart.tv_sec) +              /* in seconds */
		((double)etstart.tv_usec) / 1000000.0;  /* in microseconds */
	return now_time;
}

//Comparator to sort fibers according to the clusters of each point
bool cmp(const int *a,const int *b){
	for(int i=0;i<21;i++){
		if(Kp[i]){
			if(a[i]==b[i]) continue;
			else return a[i]<b[i]; 
		}
	}
}

//Comparator to verify if two fibers are the same in terms of the cluster values of the used points
bool cmpeq(const int *a,const int *b){
	bool eq = true;
	for(int i=0;i<21;i++){
		if(Kp[i]){
			if(a[i]==b[i]) continue;
			else {
				eq = false;
				break;
			} 
		}
	}
	return eq;
}


//Read bundles file and store it into a matrix pointHost
int read_bundle(char* bunfile, float *pointsHost[21]){
	int i, j;
	FILE *fb;
	fb = fopen(bunfile, "r");
	if (fb == NULL) {fputs ("File error",stderr); exit (1);}

	char buffer[100];   //buffer para la linea
	fgets(buffer,100,fb);//Lecturas para llegar la liena 5 del archivo
	fgets(buffer,100,fb);
	fgets(buffer,100,fb);
	fgets(buffer,100,fb);
	fgets(buffer,100,fb);
	int index_i=0;//Indice posición inicial de los numeros
	for(i=0;i<sizeof(buffer);i++){
		if(buffer[i]>=48 && buffer[i]<=57){//Detecta cuando hay un número en la línea leída y guarda el índice
			index_i=i;
			break;
		}
	}
	int index_f=index_i;//Indice posición final de los números
	int32_t nfibers=0;   // variable para numero de fibras
	while(buffer[index_f]!=44)//Mientras no se detecte una coma aumenta el otro indice de la pos final
		{
			index_f++;
		}
	for(i=0;i<(index_f-index_i);i++)
		{
			nfibers+=(buffer[index_i+i]-48)*pow(10,(index_f-(index_i+i))-1);
		}
	fclose(fb);


	int cont=0;
	char data[5] = "data";

	char* bunfileb= (char*)malloc(sizeof(char)*(strlen(bunfile)+strlen(data)));

	while(bunfile[cont]!='\0'){
		bunfileb[cont]=bunfile[cont];//Iguala string excepto el nulo
		cont++;
	}
	for(i=0;i<sizeof(data);i++){
		bunfileb[cont]=data[i];//Le agrega data al final
		cont++;
	}

	FILE *fp;
	fp = fopen(bunfileb, "rb");
	if (fp == NULL) {fputs ("File error",stderr); exit (1);}

	float** points; // puntero a cada fibra
	points = (float**) malloc (nfibers*sizeof(float*));


	for(int i=0;i<21;i++) pointsHost[i] = (float*) malloc(nfibers*3*sizeof(float));

	int32_t* npoints;//Puntero para la cantidad de puntos de cada fibra
	npoints=(int32_t*) malloc(nfibers*sizeof(int32_t));//Memoria para cada fibra

	int count = 0;
	for ( i = 0; i < nfibers; i++ )//Itera en cantidad de fibras
		{
			fread(npoints+i,sizeof(int32_t),1,fp);//Lee primer elemento(cantidad de funtos)

			points[i]=(float*) malloc((*(npoints+i))*3*sizeof(float));//Asigna memoria para toda una fibra
			if (points[i] == NULL) {fputs ("Memory error",stderr); exit (2);}
			fread( points[i], sizeof(float), *(npoints+i)*3, fp );//Lee todos los puntos de la fibra

			 
			for(int j=0;j<21*3;j++){
				pointsHost[j/3][j%3 + i*3] = points[i][j];
			}
			count+=3;
		}

	fclose(fp);
	free(bunfileb);
	return nfibers;  
}

//Write output in a file.bundle. Not separated, all belong to a unique cluster.
void write_bundle(const char* bunfile, int32_t nfibers, float *pointsHost){
	char par1[] = "attributes = {\n    'binary' : 1,\n    'bundles' : ['fibers', 0],\n    'byte_order' : 'DCBA',\n    'curves_count' : ";
	char par2[] = ",\n    'data_file_name' : '*.bundlesdata',\n    'format' : 'bundles_1.0',\n    'space_dimension' : 3\n  }";

	int32_t len=strlen(to_string(nfibers).c_str())+1;

	FILE *fw;
	fw = fopen(bunfile,"w");

	fwrite(par1,sizeof(char),strlen(par1),fw);
	fwrite(to_string(nfibers).c_str(),sizeof(char),len,fw);
	fseek(fw, -1 , SEEK_CUR);
	fwrite(par2,sizeof(char),strlen(par2),fw);

	fclose(fw);


	int cont=0,i;
	char data[5]="data";
	char* bunfileb = (char*)malloc(sizeof(char)*(strlen(bunfile)+strlen(data)+1));

	while(bunfile[cont]!='\0')
		{
			bunfileb[cont]=bunfile[cont];//Iguala string excepto el nulo
			cont++;
		}
	for(i=0;i<sizeof(data);i++)
		{
			bunfileb[cont]=data[i];//Le agrega data al final
			cont++;
		}
		

	FILE *fwb;
	fwb = fopen(bunfileb,"wb");

	int32_t *npoints = (int32_t*) malloc(sizeof(int32_t));
	int32_t b = 21;
	npoints = &b;


	for(i=0;i<nfibers;i++){
		float aux[63];
		for(int j=0;j<63;j++){
			aux[j] = pointsHost[i*63 + j];           
		}
		fwrite(npoints,sizeof(int32_t),1,fwb);
		fwrite(aux,sizeof(float),*(npoints)*3,fwb);  
	}
	fclose(fwb);
}

//Write output in a file .bundle. This format enables to generate a unique file with the clusters separated. Useful for visualization tools.
//In string bundle, for each cluster, it writes the number of the cluster and its initial fiber
void write_bundle(const char* bunfile, int32_t nfibers, float *pointsHost, string bundle){

	string str = "attributes = {\n    'binary' : 1,\n    'bundles' : " + bundle + ",\n    'byte_order' : 'DCBA',\n    'curves_count' : " + to_string(nfibers);
	char par1[str.size() + 1];
	for(int i=0;i<str.size() +  1;i++) par1[i] = str[i];
	char par2[] = ",\n    'data_file_name' : '*.bundlesdata',\n    'format' : 'bundles_1.0',\n    'space_dimension' : 3\n  }";

	int32_t len =strlen(to_string(nfibers).c_str());

	FILE *fw;
	fw = fopen(bunfile,"w");

	fwrite(par1,sizeof(char),strlen(par1)+1,fw);
    fseek(fw, -1 , SEEK_CUR);
	fwrite(par2,sizeof(char),strlen(par2),fw);

	fclose(fw);


	int cont=0,i;
	char data[5]="data";
	char* bunfileb = (char*)malloc(sizeof(char)*(strlen(bunfile)+strlen(data)+1));

	while(bunfile[cont]!='\0')
		{
			bunfileb[cont]=bunfile[cont];//Iguala string excepto el nulo
			cont++;
		}
	for(i=0;i<sizeof(data);i++)
		{
			bunfileb[cont]=data[i];//Le agrega data al final
			cont++;
		}
		

	FILE *fwb;
	fwb = fopen(bunfileb,"wb");

	int32_t *npoints = (int32_t*) malloc(sizeof(int32_t));
	int32_t b = 21;
	npoints = &b;


	for(i=0;i<nfibers;i++){
		float aux[63];
		for(int j=0;j<63;j++){
			aux[j] = pointsHost[i*63 + j];           
		}
		fwrite(npoints,sizeof(int32_t),1,fwb);
		fwrite(aux,sizeof(float),*(npoints)*3,fwb);  
	}
	fclose(fwb);
}


//Copy of points from Host to Device, it include cudaMallocs
void points_host_to_device(int numObjs, int numCoords, float *pointsHost[21], float *pointsDevice[21]){
	cudaStream_t pointsDeviceSTREAM[21];
	for(int i = 0; i < 21; i++){
		if(Kp[i]){
			cudaStreamCreate(&pointsDeviceSTREAM[i]);
			cudaMalloc(&pointsDevice[i],numObjs*numCoords*sizeof(float));
		}
	}
		
	for(int i = 0; i < 21; i++){
		if(Kp[i]){
			cudaMemcpyAsync(pointsDevice[i],pointsHost[i],numObjs*numCoords*sizeof(float),cudaMemcpyHostToDevice,pointsDeviceSTREAM[i]);
		}  
	}
	cudaDeviceSynchronize(); 
}

//It executes KMeans in GPU, it assignes memory to the output matrices (centroids and memberships)
void KMeans(int numObjs, int numCoords, float *pointsDevice[21], float *centroidsKMeansDevice[21], int *membershipDevice[21]){
	int *lp;
	float threshold = 0.001;
	cudaMalloc(&lp,sizeof(float));

        float *d_distancias[21];
        unsigned int *d_posiciones[21];
        float *d_aux[21];
  	unsigned int numObjects=3*numObjs;
  	unsigned int nextpow2=4194304;

	for(int i=0;i<21;i++){
		if(Kp[i]){
			cudaMalloc(&centroidsKMeansDevice[i],Kp[i]*numCoords*sizeof(float));
		}
	}  

        int loop = 0;
        double ck1 = wtime();
        int kpcs[5];
        kpcs[0] = Kp[0];
        kpcs[1] = Kp[3];
        kpcs[2] = Kp[10];
        kpcs[3] = Kp[17];
        kpcs[4] = Kp[20];
        float *pd[5], *ckd[5];
        int *md[5];
        pd[0] = pointsDevice[0];
        pd[1] = pointsDevice[3];
        pd[2] = pointsDevice[10];
        pd[3] = pointsDevice[17];
        pd[4] = pointsDevice[20];
        md[0] = membershipDevice[0];
        md[1] = membershipDevice[3];
        md[2] = membershipDevice[10];
        md[3] = membershipDevice[17];
        md[4] = membershipDevice[20];
        ckd[0] = centroidsKMeansDevice[0];
        ckd[1] = centroidsKMeansDevice[3];
        ckd[2] = centroidsKMeansDevice[10];
        ckd[3] = centroidsKMeansDevice[17];
        ckd[4] = centroidsKMeansDevice[20];

        for(int i=0; i<5; i++){
       	   float *ccc = (float *) malloc(kpcs[i]*numCoords*sizeof(float*));
       	   md[i] = cuda_kmeans(pd[i], numCoords, numObjs, kpcs[i], threshold, &loop, ccc);
       	   cudaMemcpy(ckd[i],ccc,kpcs[i]*numCoords*sizeof(float), cudaMemcpyHostToDevice);
       	   free(ccc);
	}
	membershipDevice[0] = md[0];
	membershipDevice[3] = md[1];
	membershipDevice[10] = md[2];
	membershipDevice[17] = md[3];
	membershipDevice[20] = md[4];
	double ck2 = wtime();
	//cout<<"cuda_kmeans time: "<<ck2-ck1<<endl;

	for(int i=0;i<21;i++){
   	   cudaFree(pointsDevice[i]);
	}
} 

//Distancia Euclidiana para puntos 3D
__device__ __host__ float d_E(float x1, float y1, float z1, float x2, float y2, float z2){
	return sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)+(z1-z2)*(z1-z2));
}

//Maxima Distancia Euclidiana para fibras, evalua la directa e inversa y retorna la mínima de las dos evaluaciones
//Recibe el numero de las fibras 1 y 2, además recibe el arreglo 1D de puntos
float d_ME(int f1, int f2, float *pointsHostArray){
	float maxdistancedirect = -1.0; 
	f1*=63;
	f2*=63;
	for(int l = 0;l<21;l++){
		float dist = d_E(pointsHostArray[f1 + l*3+0],pointsHostArray[f1 + l*3+1],pointsHostArray[f1 + l*3+2],
				 pointsHostArray[f2 + l*3+0],pointsHostArray[f2 + l*3+1],pointsHostArray[f2 + l*3+2]);
		if(dist>maxdistancedirect) maxdistancedirect=dist;
	}
	float maxdistanceflipped = -1.0; 
	for(int l = 0;l<21;l++){
		float dist = d_E(pointsHostArray[f1 + l*3+0],pointsHostArray[f1 + l*3+1],pointsHostArray[f1 + l*3+2],
				 pointsHostArray[f2 + (20-l)*3+0],pointsHostArray[f2 + (20-l)*3+1],pointsHostArray[f2 + (20-l)*3+2]);
		if(dist>maxdistanceflipped) maxdistanceflipped=dist;
	}
	return min(maxdistancedirect,maxdistanceflipped);
}

//Maxima Distancia Euclidiana para fibras, evalua la directa e inversa y retorna la mínima de las dos evaluaciones
//Recibe las fibras en formato de arreglo
float d_ME(float *f1, float *f2){
	float maxdistancedirect = -1.0; 
	for(int l = 0;l<63;l+=3){
		float dist = d_E(f1[l],f1[l+1],f1[l+2],
				 f2[l],f2[l+1],f2[l+2]);
		if(dist>maxdistancedirect) maxdistancedirect=dist;
	}
	float maxdistanceflipped = -1.0; 
	for(int l = 0;l<63;l+=3){
		float dist = d_E(f1[l],f1[l+1],f1[l+2],
				 f2[60-l],f2[60-l+1],f2[60 - l + 2]);
		if(dist>maxdistanceflipped) maxdistanceflipped=dist;
	}
	return min(maxdistancedirect,maxdistanceflipped);

}

//Predict de KMeans, para un punto X permite ver el cluster mas cercano a dicho punto en base al resultado del clustering de puntos del paso 1 
vector<int> predict(vector<float> &points, float* centroidsKmeans, int clusters){
	int sizePoints = points.size();
	int sizeCentroids = clusters*3; 

	vector<int> vec;
	for(int i=0;i<sizePoints;i+=3){
		int idx = 0;
		float minDistance = d_E(points[i],points[i+1],points[i+2], centroidsKmeans[0], centroidsKmeans[1], centroidsKmeans[2]);
		for(int j=0;j<sizeCentroids;j+=3){
			float aux = d_E(points[i],points[i+1],points[i+2], centroidsKmeans[j], centroidsKmeans[j+1], centroidsKmeans[j+2]);
			if(aux<minDistance) {
				minDistance = aux;
				idx = j/3;
			}
		}
		vec.push_back(idx);    
	}
	return vec;
}

//Lista los cliques maximales
list<list<int>> maximal_cliques(vector<vector<int>> &adjList){
	Algorithm *pAlgorithm = new AdjacencyListAlgorithm(adjList);
	list<list<int>> cliques;
	auto storeCliqueInList = [&cliques](list<int> const &clique) {
		list<int> aux(clique.begin(),clique.end());
		aux.sort();
		cliques.push_back(aux);
	};
	pAlgorithm->AddCallBack(storeCliqueInList);
	RunAndPrintStats(pAlgorithm, cliques, true);
	return cliques;
}

//Comparador para ordenar los cliques de menor a mayor tamaño
bool comparemintomax(const std::list<int>& first, const std::list<int>& second){
	if(first.size()==second.size()){
		auto itf = first.begin();
		auto its = second.begin(); 
		while(itf!=first.end() && its != second.end()){
			if(*itf == *its){
				++itf;
				++its;
				continue;  
			}
			if(*itf < *its) return true;
			else return false;
		}
	}
	return ( first.size() < second.size() );
}

//Comparador para ordenar los cliques de mayor a menor tamaño
bool comparemaxtomin(const std::list<int>& first, const std::list<int>& second){
	if(first.size()==second.size()){
		auto itf = first.begin();
		auto its = second.begin(); 
		while(itf!=first.end() && its != second.end()){
			if(*itf == *its){
				++itf;
				++its;
				continue;  
			}
			if(*itf > *its) return true;
			else return false;
		}
	}
	return ( first.size() > second.size() );
}

//Calcula la norma entre los puntos iniciales de dos fibras
__host__ __device__ float norm(float *f1, float *f2){
	return sqrt((f1[0]-f2[0])*(f1[0]-f2[0]) + (f1[1]-f2[1])*(f1[1]-f2[1]) + (f1[2]-f2[2])*(f1[2]-f2[2]));
}

//Calcula la norma entre un punto iniciale y un extremo de dos fibras
__host__ __device__ float norm_flipped(float *f1, float *f2){
	return sqrt((f1[0]-f2[60])*(f1[0]-f2[60]) + (f1[1]-f2[61])*(f1[1]-f2[61]) + (f1[2]-f2[62])*(f1[2]-f2[62]));
}

//Comprueba si una fibra esta invertida con respecto a otra
__host__ __device__ bool is_inverted(float *f1, float *f2){
	float dist_00  = norm(f1,f2);
	float dist_020 = norm_flipped(f1,f2);
	if(dist_00 > dist_020) return true;	
	else return false;
}

//Comprueba si las fibras de un cluster estan invertidas con respecto a la primera fibra de este, si alguna lo esta la rota
__host__ __device__ float* align_fibers(float *cluster, int size_cluster){
	float *new_fibers = (float*) malloc(63*size_cluster*sizeof(float));
	float *f1 = (float*) malloc(63*sizeof(float));
	float *f2 = (float*) malloc(63*sizeof(float));
	
	for(int i=0;i<63;i++){
		new_fibers[i] = cluster[i];
		f1[i]=cluster[i];
	}
	for(int i=1;i<size_cluster;i++){
		for(int j=0;j<63;j++){
			f2[j]=cluster[i*63 + j];
		}
		if(is_inverted(f1,f2)){
			for(int j=0;j<63;j++){
				new_fibers[i*63+j] = f2[62-j];
			}
		}
		else{
			for(int j=0;j<63;j++){
				new_fibers[i*63+j] = f2[j];
			}
		}
	}
	free(f1);
	free(f2);
	return new_fibers;
}

//Para un cluster computa el centroide, aplicando la media aritmetica
__host__ __device__ float* centroid_mean_align(float* cluster,int size_cluster){
	float *fibers = align_fibers(cluster,size_cluster);
	float *centroid = (float*) malloc(63*sizeof(float));

	for(int i=0;i<63;i++) centroid[i]=0.0;

	for(int i=0;i<size_cluster;i++){
		for(int j=0;j<63;j++){
			centroid[j] += fibers[i*63 + j];
		}
	}
	for(int i=0;i<63;i++) centroid[i] = centroid[i]/(float)size_cluster;
	return centroid;
}


//Computa los centroides de los cluster obtenidos
float* compute_centroids(vector<vector<int>> &fibersclusters,float *pointsHostArray){
	float *centroids = (float*) malloc(63*fibersclusters.size()*sizeof(float));
	float **clusters = (float**) malloc(fibersclusters.size()*sizeof(float*));

	for(int i=0;i<fibersclusters.size();i++){
		clusters[i] = (float*) malloc(63*fibersclusters[i].size()*sizeof(float));
	}

	//#pragma omp parallel for
	for(int i=0;i<fibersclusters.size();i++){
		for(int j=0;j<fibersclusters[i].size();j++){
			for(int k=0;k<63;k++){
				clusters[i][j*63+k] = pointsHostArray[fibersclusters[i][j]*63+k]; 
			}
		}
	}

	for(int i=0;i<fibersclusters.size();i++){
		float *centroid = centroid_mean_align(clusters[i],fibersclusters[i].size());
		for(int j=0;j<63;j++){
			centroids[i*63 + j] = centroid[j];
		}
	}
	return centroids;
}

//compute centroids final clusters
float* compute_finalcentroids(vector<vector<int>> &fibersclusters,float *pointsHostArray){
	float *centroids = (float*) malloc(63*fibersclusters.size()*sizeof(float));
	float **clusters = (float**) malloc(fibersclusters.size()*sizeof(float*));

	for(int i=0;i<fibersclusters.size();i++){
		clusters[i] = (float*) malloc(63*fibersclusters[i].size()*sizeof(float));
	}

	//#pragma omp parallel for
	for(int i=0;i<fibersclusters.size();i++){
		for(int j=0;j<fibersclusters[i].size();j++){
			for(int k=0;k<63;k++){
				clusters[i][j*63+k] = pointsHostArray[fibersclusters[i][j]*63+k]; 
			}
		}
	}

	for(int i=0;i<fibersclusters.size();i++){
		float *centroid = (float*) malloc(63*sizeof(float));
		//centroid = centroid_mean_align(clusters[i],fibersclusters[i].size());
		int size_cluster = fibersclusters[i].size();

	        for(int j=0;j<63;j++) centroid[j]=0.0;

       		for(int j=0;j<size_cluster;j++){
                   for(int k=0;k<63;k++){
                        centroid[k] += clusters[i][j*63+k];
                   }
        	}
        	for(int j=0;j<63;j++) centroid[j] = centroid[j]/(float)size_cluster;

		//for(int j=0;j<63;j++) centroid[j] = clusters[i][j];
		for(int j=0;j<63;j++){
			centroids[i*63 + j] = centroid[j];
		}
		
	}
	return centroids;
}

// Creates graphs and compute maximal cliques
void join_cliques(vector<vector<int>> &atlas, float *centroidsKMeansHost,float *pointsHostArray){
	float *centroids = compute_centroids(atlas,pointsHostArray);

	vector<float> centroidesvector;
	for(int i=0;i<atlas.size();i++){
		for(int j=0;j<3;j++){
			centroidesvector.push_back(centroids[i*63 + j + 30]);
		}
	}

	vector<int> membership_centroid = predict(centroidesvector,centroidsKMeansHost,Kp[10]);
	vector<vector<int>> groups(Kp[10]);

	for(int i=0;i<atlas.size();i++){
		groups[membership_centroid[i]].push_back(i);
	}

	vector<bool> visited(atlas.size(),false);

#pragma omp parallel
{
	#pragma omp for
	for(int i=0;i<Kp[10];i++){
		vector<vector<int>> graph(groups[i].size());
		float *f1 = (float*) malloc(63*sizeof(float));
		float *f2 = (float*) malloc(63*sizeof(float));  

		for(int j=0;j<groups[i].size();j++){
			for(int k=j+1;k<groups[i].size();k++){
				for(int l = 0; l<63; l++){
					f1[l] = centroids[groups[i][j]*63 + l];
					f2[l] = centroids[groups[i][k]*63 + l];
				}
				float maxdistance = d_ME(f1,f2);

				//if(maxdistance<=7.0){
				if(maxdistance<=dMmax){ // Ceci test con dRmax, dMMax 7
					graph[j].push_back(k);
					graph[k].push_back(j);
				}
			}
		}

		list<list<int>> cliques = maximal_cliques(graph);

                //cliques.sort(comparemaxtomin);

                auto cmp = [](std::vector<int> left, std::vector<int> right) { return(left.size() < right.size());};

                std::priority_queue<std::vector<std::vector<int>>, std::vector<std::vector<int>>, decltype(cmp)> q(cmp);
                list<list<int>>::iterator ll;

                for (ll=cliques.begin(); ll!=cliques.end(); ll++) {
                        list<int> ml = *ll;
                        vector<int> r(ml.begin(), ml.end());
                        q.push(r);
                }


                while(!q.empty()){
                        vector<int> aux = q.top();      
                        q.pop();
			if(aux.size()>=2){
				int first = -1;
				bool joined = false; 
				for(auto it = aux.begin(); it!=aux.end();++it){
					int a = groups[i][*it];         
					if(first == -1 && visited[a]==false) {
						first = a;
					}
					else{
						bool flag;
						//#pragma omp critical
						flag = visited[a];
						if(!flag){
							//#pragma omp critical
							for(int i = 0;i<atlas[a].size();i++){
								if(visited[a]==false && visited[first]==false) {
									atlas[first].push_back(atlas[a][i]);
									joined = true;
									//puts("si hay cliques, entonces entro");
								}
							}
							visited[a] = true;
							if(atlas[a].size()>0) atlas[a].clear();
						}         
					}
				}
				if(joined) visited[first] = true;
			}
		}
	}
}
}

int main(int argc, char *argv[]) {
        if(argc < 8){
		cout<<"usage: "<<argv[0]<<" path_to_files_bundles_and_bundledata output_dir cluster_size_min cluster_size_max cluster_size_step dRmax dMmax\n";
		cout<<" example:\n";
		cout<<" optirun ./ffclustgpu /home/chernand/udec/Alltesis/IsaacGoicovich/v2/subs_limpios/subs_limpios/sub0.bundles output 100 2000 100 6 6 \n";
		return 1;
	}
	init();
	cudaDeviceReset();
	double start,end;
	double tiempoacumulado = 0.0;

        int x_init = atoi(argv[3]);
        int x_end = atoi(argv[4]);
        int x_step= atoi(argv[5]);
	dRmax = atof(argv[6]);
	dMmax = atof(argv[7]);

	int   numObjs, numCoords = 3;
	float *pointsHost[21];
	float *pointsHostArray;

	float *pointsDevice[21];
	float *pointsDeviceArray;

	float *centroidsKMeansDevice[21];
	float *centroidsKMeansHost[21]; 
	int *membershipDevice[21];
	int *membershipHost[21];  
	int **membershipHostTrasposed; 
	int *key;
	int *B;
	int **C;
	int *D;
 

 
	//Lectura de los datos de entrada y creación de directorios, los almacena en Host y Device
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	puts("Start reading ");
	start = wtime();

	for(int i=0;i<numObjs;i++){
		membershipHostTrasposed[i] = (int*) malloc(21 * sizeof(int));
		key[i]=i;
	}

	numObjs = read_bundle(argv[1],pointsHost);
	//cout<<numObjs<<" fibers\n";
	points_host_to_device(numObjs, numCoords,pointsHost,pointsDevice);
	
	pointsHostArray = (float*) malloc(63*numObjs*sizeof(float));
	cudaMalloc(&pointsDeviceArray, sizeof(float) * numObjs*63);
	

	#pragma omp parallel for
	for(int i=0;i<numObjs;i++){
		for(int j=0;j<63;j++){
			pointsHostArray[i*63+j] = pointsHost[j/3][j%3 + i*3];
		}
	}
	cudaMemcpy(pointsDeviceArray,pointsHostArray,63*numObjs*sizeof(float),cudaMemcpyHostToDevice);

	membershipHostTrasposed = (int**) malloc(numObjs*sizeof(int*));
	key = (int*) malloc(numObjs*sizeof(int));
	B = (int*) malloc(numObjs * sizeof(int));
	C = (int**) malloc(numObjs * sizeof(int*));
	D = (int*) malloc(numObjs * sizeof(int));

	for(int i=0;i<numObjs;i++){
		membershipHostTrasposed[i] = (int*) malloc(21 * sizeof(int));
		key[i]=i;
		B[i]=1;
	}


	//Creation of output directory
	mkdir(argv[2],0777);
	string dirOutput(argv[2]);
	//string dir = dirOutput + "/bundlesPerCluster";
	//mkdir(dir.c_str(),0777);
	string dir = dirOutput + "/bundlesPerSize";
	mkdir(dir.c_str(),0777);
	dir = dirOutput + "/bundleFinalClustered";
	mkdir(dir.c_str(),0777);
	dir = dirOutput + "/bundleCentroid";
	mkdir(dir.c_str(),0777);


	//Creation of output stats.log 
	FILE *log;
	string stats(argv[2]);
	stats = stats + "/stats.log";
	log = fopen(stats.c_str(),"w");

	end = wtime();
	cout<<"Reading and copy time of data to device: "<<end-start<<endl;
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


	//Step 1: Building point clusters. It call KMeans on fiber points
	//
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

        cudaEvent_t stt, stp;
        cudaEventCreate(&stt);
        cudaEventCreate(&stp);
	start = wtime();  
	puts("Step 1: Building point clusters");
	cudaEventRecord(stt);
	for(int i=0;i<21;i++){
	   membershipHost[i] = (int*) malloc(numObjs * sizeof(int));
	   centroidsKMeansHost[i] = (float*) malloc(Kp[i]*numCoords*sizeof(float));
	}
	KMeans(numObjs,numCoords,pointsHost,centroidsKMeansDevice,membershipDevice);

	for(int i=0;i<21;i++){
		cudaMemcpy(membershipHost[i],membershipDevice[i],numObjs*sizeof(float),cudaMemcpyDeviceToHost);   
		cudaMemcpy(centroidsKMeansHost[i],centroidsKMeansDevice[i],Kp[i]*numCoords*sizeof(float),cudaMemcpyDeviceToHost);
		cudaFree(membershipDevice[i]);
	}

	#pragma omp parallel for
	for(int i=0;i<numObjs;i++){
		for(int j=0;j<21;j++){
			membershipHostTrasposed[i][j] = membershipHost[j][i];
		}
	}
	end = wtime();
	tiempoacumulado += end-start; 
	printf("Step 1 time: %lf\n",end-start);
	fprintf(log,"Step 1 time: %lf\n",end-start);
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//Step 2: Building preliminary clusters
	//
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	//puts("Empiezo mapeo");
	start = wtime();
	thrust::sort_by_key(thrust::host,membershipHostTrasposed, membershipHostTrasposed + numObjs,key,cmp);
	thrust::pair<int**,int*> new_end;
	new_end = thrust::reduce_by_key(thrust::host,membershipHostTrasposed, membershipHostTrasposed + numObjs,B,C,D,cmpeq);
	end = wtime();
	tiempoacumulado += end-start;
	printf("Step 2 time: %lf\n",end-start);
	fprintf(log,"Step 2 time: %lf\n",end-start);

	//Step 3: Reassignment of small preliminary clusters to large
	//        
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////  

	//Computing centroids
	//puts("Calculando centroides en CPU");
	start = wtime();

	vector<vector<int>> atlas;
	vector<vector<int>> sujeto;
	int size_clusters = new_end.first - C;
	int start2 = 0;
	int maxsizecluster = 0;
	for(int i = 0;i<size_clusters;i++){
		vector<int> cluster;
		for(int j=0;j<D[i];j++){
			cluster.push_back(key[j+start2]);
		}
		start2+=D[i];
		if(cluster.size()<6) sujeto.push_back(cluster);
		else atlas.push_back(cluster);
		if(cluster.size()>maxsizecluster) maxsizecluster = cluster.size();
	}


	std::sort(atlas.begin(), atlas.end(),
			[] (const auto& lhs, const auto& rhs) {
				return lhs.size() > rhs.size();
			});

	//puts("Calculando centroides atlas");
	float *ATLAS = compute_centroids(atlas,pointsHostArray);
	//puts("Calculando centroides sujeto");
	float *SUJETO = compute_centroids(sujeto,pointsHostArray);
	
	printf("Total Ss clusters with 5 fibers or fewer:  %d\n",sujeto.size());
	printf("Total SL clusters with 6 fibers or more: %d\n",atlas.size());

	// Reassignment
	int *assignment = segmentation(21,SUJETO,ATLAS,dRmax,sujeto.size(),atlas.size());
        int clusters2Fibras = 0;
        int clusters1Fibras = 0;

	int num_fibers_reassigned = 0;
	int num_fibers_discarded  = 0;
	for(int i=0;i<sujeto.size();i++){
		if(assignment[i]!=-1){
			for(int j=0; j<sujeto[i].size();j++){
				atlas[assignment[i]].push_back(sujeto[i][j]);
				num_fibers_reassigned++;
			}
		}
		else if(sujeto[i].size() > 2){ 
			atlas.push_back(sujeto[i]);
		}
		else num_fibers_discarded++;
	}

	printf("Total candidate clusters: %d\n",atlas.size());
	printf("Total discarded clusters: %d\n",num_fibers_discarded);
	printf("Total reassigned clusters: %d\n",num_fibers_reassigned);
	
	end = wtime();
	tiempoacumulado += end-start;
	printf("Step 3 time: %f\n",end-start);
	fprintf(log,"Step 3 time: %lf\n",end-start);
	
	//Step 4: Final clusters using graph representation 
	//        
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	start = wtime();

	join_cliques(atlas,centroidsKMeansHost[10],pointsHostArray);
	
	int sizeaux = atlas.size();
	for(int i=sizeaux-1;i>=0;i--){
		if(atlas[i].size()==0) atlas.erase(atlas.begin() + i);
	}

	end = wtime();
	printf("Step 4 time: %f\n",end-start);
	fprintf(log,"Step 4 time: %lf\n\n",end-start);
	tiempoacumulado += end-start;
 
	int num_covered_fibers = 0;
	for(int i=0;i<atlas.size();i++){
		num_covered_fibers+=atlas[i].size();
	}
	printf("Total fibers covered: %d\n",num_covered_fibers);
	printf("Total number of final clusters: %d\n",atlas.size());
	cudaDeviceSynchronize();


	//output, files clusterPerSize.bundles clusters of sizes between xinit and xend with step [X,X+50)
	//
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// 

	for(int X = x_init; X<=x_end; X+=x_step){
		float *fibers_per_cluster = (float*) malloc(63*num_covered_fibers*sizeof(float));
		string str = "[";
	  	int32_t number_of_fibers = 0;
	  	int cluster_number = 0;
	  	int first_fiber_cluster = 0;
		int aux_count = 0;
		bool exist = false;

		for(int i=0;i<atlas.size();i++){
			if(atlas[i].size()>=X && atlas[i].size()<X+x_step){
			//if(atlas[i].size()==X) {
				string aux;
				aux += "'" + to_string(cluster_number) + "', "+ std::to_string(first_fiber_cluster) + ", ";
				first_fiber_cluster+=atlas[i].size();

				for(int j=0;j<atlas[i].size();j++){    
		  			for(int k=0;k<63;k++){
						fibers_per_cluster[aux_count++] = pointsHostArray[atlas[i][j]*63 + k];
					}
		  			number_of_fibers++;
				}
				cluster_number++;
		  		str = str + aux; 
				exist = true;
	  		}   	
		}

		if(!exist)continue;
		str.pop_back();
		str.pop_back();
		str = str + "]";

		string auxaux(argv[2]);
		
		string pathout = auxaux  + "/bundlesPerSize/clustersize"+ to_string(X)+ ".bundles";
		write_bundle(pathout.c_str(),number_of_fibers,fibers_per_cluster,str);
		fprintf(log,"Number of fibers in clust%d: %d\n",X, number_of_fibers);
		free(fibers_per_cluster);

	}
	fprintf(log,"\n");
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	//computing centroids of final clusters
	float *clusterfinales_device = compute_finalcentroids(atlas,pointsHostArray);
	start = wtime();
	int32_t obj = atlas.size();
	string output_directory(argv[2]);
	output_directory = output_directory + "/bundleCentroid/centroids.bundles";
	write_bundle(output_directory.c_str(), obj, clusterfinales_device);

	//output, files centroids.bundles of final clusters
	//
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////// 

	float *fibers_per_cluster = (float*) malloc(63*num_covered_fibers*sizeof(float));
	string str = "[";
  	int32_t number_of_fibers = 0;
  	int cluster_number = 0;
  	int first_fiber_cluster = 0;
	int aux_count = 0;

	for(int i=0;i<atlas.size();i++){
		string aux;
		if(atlas.size() <= 2)continue; // does not output clusters with 2 or less fibers
		aux += "'" + to_string(cluster_number) + "', "+ std::to_string(first_fiber_cluster) + ", ";
		first_fiber_cluster+=atlas[i].size();

		for(int j=0;j<atlas[i].size();j++){    
  			for(int k=0;k<63;k++){
				fibers_per_cluster[aux_count++] = pointsHostArray[atlas[i][j]*63 + k];
			}
  			number_of_fibers++;
		}
		cluster_number++;
  		str = str + aux; 
	}
	str.pop_back();
	str.pop_back();
	str = str + "]";

	string auxaux(argv[2]);
	string pathout = auxaux  + "/bundleFinalClustered/clustered.bundles";
	write_bundle(pathout.c_str(),number_of_fibers,fibers_per_cluster,str);
	free(fibers_per_cluster);
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


	//free memory
	///////////////////////////
	
	for(int i=0;i<21;i++){
		cudaFree(pointsDevice[i]);
		cudaFree(centroidsKMeansDevice[i]);
		cudaFree(membershipDevice[i]);
		free(pointsHost[i]);
	}
	cudaDeviceSynchronize();
	end = wtime();

	printf("Done writing output files: %lf\n",end-start);

	puts("");
	printf("Total final clusters: %d\n",atlas.size());
	printf("TOTAL TIME: %lf\n", tiempoacumulado);
	fprintf(log,"TOTAL TIME: %lf\n", tiempoacumulado);
	fprintf(log,"Total final clusters: %d\n",atlas.size());
	fprintf(log,"Number of input fibers: %d\n",numObjs);
	fprintf(log,"Covered fibers: %d\n",num_covered_fibers);
	fprintf(log,"Percentage of covered fibers: %d\n",(num_covered_fibers*100)/numObjs);

	fclose(log);
	return 0 ;
}
